/*
	Server running on a NuBus card
	
	Written by:	Anumele D. Raja
	
	Date:		February 23, 1990
	
	Copyright  1990, Apple Computer, Inc.  All rights reserved.
	
	This example is to show the users the use of message passing
		mechanism to pass small amount of information and as an
		event delivery mechanism.
		
	This a task on the NuBus card that acts as a server.
	
		This task registers itself with:
		
			object name		-	"CodeAnalyzer"
			type name		-	"Server"

		This task can accept following codes
		
		If the buffer belongs to the task sending the message
		
		CA_COUNT		= 0x00	Count the number of distinct words in the buffer
		CA_SORT			= 0x02	Sort the words in the buffer
		
		If the buffer belongs to someother remote task whose ID
			is in mOData[2]
			
		CA_COUNT_REM	= 0x04	Count the number of distinct words in the buffer
		CA_SORT_REM		= 0x06	Sort the words in the buffer

		CA_STOP			= 0x80	Quit
		
		A message is sent to the server task with the address and
		the length of the buffer
		
		It returns in
		
		mOData[0]	-	Number of distict words in the buffer
		
		For all messages this program replies with the mCode
			incremented by one. mOData[2] remains unchanged.
						
	Note:
	
		This task is downloaded as a dynamic task.
		This program is linked with SAServer to generate an intial load image.
*/

#include <Types.h>
#include <OS.h>
#include <Managers.h>

#include "ServerClient.h"

pascal illegal() = { 0x4AFC };

// Local routines

Boolean GetSenderBuffer(char *localBuff, AROSEmessage *msgPtr);
Boolean GetRemoteBuffer(char *localBuff, AROSEmessage *msgPtr);
void PutSenderBuffer(char *localBuff, AROSEmessage *msgPtr, long buffLen);
void PutRemoteBuffer(char *localBuff, AROSEmessage *msgPtr, long buffLen);
void SortBuffer(char *localBuff, long buffLen);
long CountDistinctWords(char *localBuff, long buffLen);

#pragma segment "Server"

server()
{
	AROSEmessage *msgPtr;
	char *localBuffPtr;
	long buffLen;
	Boolean done;

	//  Register my name with the Name Manger
	
	printf("Server registering on card %X\n", GetCard());
	
	if (Register_Task(serverObjName, serverTypName, Machine_Visible)) {
		
		done = kFalse;
		
		while (!done) {
			// Wait for and get a command message
			
			msgPtr = Receive(OS_MATCH_ALL, OS_MATCH_ALL, OS_MATCH_ALL, OS_NO_TIMEOUT);
			
			printf("Server got a message with mCode %d Data Size %d\n", msgPtr->mCode, msgPtr->mDataSize);
			
			buffLen = msgPtr->mDataSize;

			if ( (localBuffPtr = GetMem(buffLen)) != 0) {
				msgPtr->mStatus = 0;	// assume everything OK
				switch (msgPtr->mCode) {
					case CA_COUNT:
						if (GetSenderBuffer(localBuffPtr, msgPtr) == kTrue) {
							SortBuffer(localBuffPtr, buffLen);
							msgPtr->mOData[0] = CountDistinctWords(localBuffPtr, buffLen);
						} else
							msgPtr->mStatus = 0x8000;
						break;
					case CA_SORT:
						if (GetSenderBuffer(localBuffPtr, msgPtr) == kTrue) {
							SortBuffer(localBuffPtr, buffLen);
							msgPtr->mOData[0] = CountDistinctWords(localBuffPtr, buffLen);
							PutSenderBuffer(localBuffPtr, msgPtr, buffLen);
						} else
							msgPtr->mStatus = 0x8000;
						break;
					case CA_COUNT_REM:
						if (GetRemoteBuffer(localBuffPtr, msgPtr) == kTrue) {
							SortBuffer(localBuffPtr, buffLen);
							msgPtr->mOData[0] = CountDistinctWords(localBuffPtr, buffLen);
						} else
							msgPtr->mStatus = 0x8000;
						break;
					case CA_SORT_REM:
						if (GetRemoteBuffer(localBuffPtr, msgPtr) == kTrue) {
							SortBuffer(localBuffPtr, buffLen);
							msgPtr->mOData[0] = CountDistinctWords(localBuffPtr, buffLen);
							PutRemoteBuffer(localBuffPtr, msgPtr, buffLen);
						} else
							msgPtr->mStatus = 0x8000;
						break;
					case CA_STOP:
						done = kTrue;
						break;
					default:
						break;
				}
				FreeMem(localBuffPtr);	// Free message
			}
			else {
			
				// the mStatus field can be used to indicate error
				
				msgPtr->mStatus = 0x8000;
			}
			
			//  Reply to the client
			
			SwapTID(msgPtr);		// interchange mFrom and mTo
			msgPtr->mCode += 1;		// prepare reply code
			Send(msgPtr);
		}
	}
	else {
		printf("Could not Register Task\n");
	}
}

// Get buffer from sender.  Returns kTrue if successful else returns kFals

#pragma segment "Server Support"

Boolean GetSenderBuffer(char *localBuff, AROSEmessage *msgPtr)
{
	if (NetCopy(msgPtr->mFrom, msgPtr->mDataPtr, GetTID(), localBuff, msgPtr->mDataSize) == 0)
		return kTrue;
	else
		return kFalse;
}

// Get buffer from remote task.  Returns kTrue if successful else returns kFalse

Boolean GetRemoteBuffer(char *localBuff, AROSEmessage *msgPtr)
{
	if (NetCopy(msgPtr->mOData[2], msgPtr->mDataPtr, GetTID(), localBuff, msgPtr->mDataSize) == 0)
		return kTrue;
	else
		return kFalse;
}

// Put buffer into sender space

void PutSenderBuffer(char *localBuff, AROSEmessage *msgPtr, long buffLen)
{
	NetCopy(GetTID(), localBuff, msgPtr->mFrom, msgPtr->mDataPtr, buffLen);
}

// Put buffer to remote task space

void PutRemoteBuffer(char *localBuff, AROSEmessage *msgPtr, long buffLen)
{
	NetCopy(GetTID(), localBuff, msgPtr->mOData[2], msgPtr->mDataPtr, buffLen);
}

// Shell sort procedure

void SortBuffer(char *localBuff, long buffLen)
{
	unsigned short *buffPtr, temp;
    int gap, i, j;
	
	// We are going to sort by words
	
	buffLen /= 2;	// number of words
	buffPtr = (unsigned short *)localBuff;
	
    for (gap = buffLen/2; gap > 0; gap /= 2) {
		for (i = gap; i < buffLen; i++) {
			for (j = i-gap; j >= 0; j -= gap) {
				if (buffPtr[j] <= buffPtr[j+gap])
					break;
				temp = buffPtr[j];
				buffPtr[j] = buffPtr[j+gap];
				buffPtr[j+gap] = temp;
			}
		}
	}
}

// Count distinct words in the buffer

long CountDistinctWords(char *localBuff, long buffLen)
{
	long wordCount, i;
	unsigned short word, *buffPtr;
	
	buffPtr = (unsigned short *)localBuff;
	wordCount = 1;
	word = *buffPtr;
	
	for (i = 0; i < buffLen/2; i++) {
		if (word != buffPtr[i]) {
			wordCount++;
			word = buffPtr[i];
		}
	}
	
	return wordCount;
}
