/********************************************************************************/
/*																				*/
/*		ciccm.c - "C" version of ICC Manager.									*/
/*																				*/
/*		February 20, 1991.  Richard W. Mincher and Richard H. Sackett.			*/
/*		Based on original 68000 version by Mark Rustad.							*/
/*																				*/
/*		Copyright  1991, Apple Computer, Inc.  All rights reserved.			*/
/*																				*/
/********************************************************************************/


#include "arose.h"
#include "os.h"
#include "iccmdefs.h"
#include "machineDep.h"
#include "types.h"
#include "managers.h"


struct iccmGlobals
{
	void	(*iTick)();			/*	Pointer to next tick chain routine	*/
	void	(*iIdle)();			/*	Pointer to next idle chain routine	*/
	long	iBusErr;			/*	Saved bus error vector				*/
	long	iCont;				/*	Where to continue from bus error	*/

	message	*iHeads[ic_SlotMax];/*	Outbound queue head pointers		*/
	message	*iTails[ic_SlotMax];/*	Outbound queue tail pointers		*/

	struct	ca_Rec	iCA;		/*	Storage for communications area		*/

	message	*iRdyMsg;			/*	Holds pointer to slot ready message	*/
	message *iRMP;				/*	Holds original pointer to slot ready message	*/

	long	iQCount;			/*	Number of times message was queued	*/
	long	iRCount;			/*	Messages returned					*/
	long	iICount;			/*	Messages received					*/
	long	iOCount;			/*	Messages sent						*/

	long	QSCount[ic_SlotMax];/*	Last QSCount value from active slot	*/
	long	Timer;				/*	Next time to check QSCount values for change	*/
	long	TimerInc;			/*	Time increment between checking for active slots	*/

	short	iWaitMask;			/*	Mask of slots being waited for		*/
	short	iSlot;				/*	My slot number						*/
	short	iSlot4;				/*	My slot number * 4					*/
	char	iInScan;			/*	Denote critical region in scanner	*/
	char	iHigh;				/*	High byte of my minor nubus space	*/
	long	SIntPtr[ic_SlotMax];					/* Set Interrupt routine parameter	*/
	long	SIntRtn[ic_SlotMax*ic_RtnSize];		/*Set Interrupt routine code	*/
	char	scanagn;			/*	Set 0 at beginning of Scanner, Set non-zero if intercard interrupt	*/
	char	inScanFlag;			/*	set, if in scanner	*/
};


void	ICCM_Idle( void );
void	ICCM_Tick( void );
void	Q2Slot( message * );
void	POQ(void);
void	Scanner( void );
void	InitCA(short);
void	ICCM_Init( void );
void	RCI( message * );
void	FindCards( void );
void	IntCard( short );
void	Scan( short );

struct	iccmGlobals iccmGlob;

/********************************************************************************/
/*																				*/
/*		ICCM - main routine for ICC manager.									*/
/*																				*/
/*		This routine processes incoming messages and requests for service.		*/
/*																				*/
/*		Entry	None.															*/
/*																				*/
/*		Exit	None.															*/
/*																				*/
/********************************************************************************/

ICCM()
{
	register struct	iccmGlobals *iccmG;
	message	*msg;
	short	i;
	
	iccmG = &iccmGlob;
	ICCM_Init();
	
	while(true)
	{
		if (iccmG->iCA.IFlag)
			Scanner();
		
		msg = Receive( OS_MATCH_ALL, OS_MATCH_ALL, OS_MATCH_ALL, nil );

/*	Check for no message received.	*/

		if (msg == nil)
			continue;
		
/*	See if it's for this card.	*/
		
		if (((msg->mTo >> 24) & 0xFF) != iccmG->iSlot)
		{
			Q2Slot( msg );
			continue;
		}
		
/*	Process message for us.	*/

		switch( msg->mCode )
		{
			case ICC_GETCARDS:
			case ICC_GETIMMS:
				RCI(msg);								/*	Return card information	*/
				break;
			case ICC_DETACH:
				iccmG->iCA.Flag = nil;					/*	Clear magic flag	*/
				for(i=0; i < ic_SlotMax; i++)
					iccmG->iCA.Addrs[i] = nil;
				gCom->gMainICCM = nil;
				SwapTID( msg );
				msg->mCode |= 1;
				Send( msg );
				break;
			case ICC_ATTACH:
				InitCA(false);						/*	Modified InitCA	*/
				FindCards();						/*	Find all cards	*/
				SwapTID( msg );
				msg->mCode |= 1;
				Send( msg );
				break;
			case ICC_INTCARD:
				SwapTID( msg );
				msg->mCode |= 1;					/*	make reply	*/
				msg->mOData[1] = iccmG->iCA.IntFlg;	/*	Get previous interrupt flag	*/
				if (msg->mOData[0])
					iccmG->iCA.IntFlg = 0xFF;
				else
					iccmG->iCA.IntFlg = 0x01;
				Send( msg );
				break;
			case ic_Rdy:
				if (msg == iccmG->iRMP)				/*	if it's our message	*/
				{
					POQ();
					iccmG->iRdyMsg = msg;
					break;
				}
			default:
				if (msg->mCode & 0x8000)
					FreeMsg( msg );
				else
				{
					msg->mCode |= 0x8000;
					msg->mStatus = OS_UNKNOWN_MESSAGE;
					SwapTID( msg );
					Send( msg );
					iccmG->iRCount++;
				}
		}
	}
}
		
/********************************************************************************/
/*																				*/
/*		Scanner - check for newly arrived inter-card messages.					*/
/*																				*/
/*		This routine is called from either the idle call chain, the major tick 	*/
/*		call chain or the Inter-card interrupt routine.  After checking the		*/
/*		scanner interlock it does the following:								*/
/*			Checks for any requests from other slots.							*/
/*			Checks to see if any cards have died.								*/
/*			Checks for outbound messages.										*/
/*																				*/
/*		Entry	Global area set up (A5).										*/
/*																				*/
/*		Exit	iRdyMsg sent to mainline code if work to do.					*/
/*																				*/
/********************************************************************************/

void 
Scanner()
{
	register struct	iccmGlobals *iccmG;
	short			slot;
	struct ca_Rec	*OtherCA;		/* Storage for communication area */
		
	iccmG = &iccmGlob;
	if (!interlock(&iccmG->inScanFlag))	/*	if already in scanner	*/
		return;
	do
	{
		iccmG->scanagn = nil;
		iccmG->iCA.NameMgr = GetNameTID();
		iccmG->iCA.IMMTID = GetICCTID();
		iccmG->iCA.gDebugOn = GetgCommon()->gDebugOn;
		iccmG->iCA.QSCount++;
		
		if(iccmG->iCA.IFlag)	/*	If any request from another slot	*/
		{			
			iccmG->iCA.IFlag = 0;
			for(slot = 0; slot < ic_SlotMax; ++slot)
				if(iccmG->iCA.IFlags[slot])		/*	If work request from this slot	*/
					Scan(slot);
		}
		
/*		See if we should do slot timeouts	*/

		if(GetgCommon() -> gMajorTick >= iccmG->Timer)
		{
			iccmG->Timer += iccmG->TimerInc;
			
			for(slot = 0; slot < ic_SlotMax; ++slot)
			{
				if(iccmG->iCA.Addrs[slot])		/*	If slot is active	*/
				{
					OtherCA = (struct ca_Rec *)MapNuBus((char *)iccmG->iCA.Addrs[slot]);
					if(OtherCA->Flag == ~ic_FVal)		/*	If other slot has not gone away	*/
					{						
						if((OtherCA->QSCount != iccmG->QSCount[slot]) || OtherCA->gDebugOn)	/*	If slot has not timed out or cannot (because of gDebugOn)	*/
						{							
							iccmG->QSCount[slot] = OtherCA -> QSCount;
							OtherCA->Addrs[iccmG->iSlot] = GetgCommon() -> gCAP.iPointer;
						}
						else
							iccmG->iCA.Addrs[slot] = 0;	
					}
					else
						iccmG->iCA.Addrs[slot] = 0;		/*	Other slot has gone away				*/
				}
			}
		}
		
/*		Check for outbound messages waiting	*/
		
		if(iccmG->iWaitMask && iccmG->iRdyMsg)		/*	If messages waiting	*/
		{			
			for(slot = 0; (slot < ic_SlotMax) && iccmG->iRdyMsg; ++slot)
			{
				if(iccmG->iWaitMask & (1 << slot))		/*	If we are waiting to send a message to this slot	*/
				{
					if(iccmG->iCA.Addrs[slot])	/*	If slot is active	*/
					{						
						OtherCA = (struct ca_Rec *)MapNuBus((char *)iccmG->iCA.Addrs[slot]);
						if(OtherCA -> Flag == ~ic_FVal)		/*	If slot is not gone away	*/
						{
							if(!OtherCA -> IFlags[iccmG->iSlot])	/*	If entries in other slots communications area belong to us	*/
							{								
								if(OtherCA -> Ptrs[iccmG->iSlot])	/*	If a message buffer present	*/
								{									
									Send(iccmG->iRdyMsg);
									iccmG->iRdyMsg = 0;
								}
								else								/*	If no message buffer present	*/
								{									
									OtherCA -> IFlags[iccmG->iSlot] = 0xff;		/*	Indicate we need buffer	*/
									OtherCA -> IFlag = 0xff;
									IntCard(slot);								/*	Interrupt slot	*/
								}
							}
						}
						else								/*	If slot has gone away	*/
						{
							Send(iccmG->iRdyMsg);
							iccmG->iRdyMsg = 0;
						}
					}
					else								/*	Slot is no longer active	*/
					{						
						Send(iccmG->iRdyMsg);
						iccmG->iRdyMsg = 0;
					}
				}
			}
		}
	}	while(iccmG->scanagn);
	
	iccmG->inScanFlag = nil;				/*	clear interlock	*/
}


/********************************************************************************/
/*																				*/
/*		FindCards - Locate other cards running ICCM.							*/
/*																				*/
/*		This routine scans other cards (and the mother board) for ICCM			*/
/*		communications areas.  These are found by looking for the magic number	*/
/*		in the first 128K bytes of memory (1 megabyte on the motherboard).		*/
/*		If a valid area is found, it is added to our list of known cards, and	*/
/*		our communications area pointer is put in their list of known cards.	*/
/*																				*/
/*		Entry	none.															*/
/*																				*/
/*		Exit	Communications area updated.									*/
/*																				*/
/********************************************************************************/


void
FindCards()
{
	short	s, slot;
	short	boardID, CPURsrc[4], networkRsrc[4];
	long	lenMem;
	char	*iccmArea, *iccmEnd;
	struct	ca_Rec	*otherCA, *rawca;
	struct	gCAP	*gcap;
	void	ReadError();
	OSlong	ReadLong();
	register struct	iccmGlobals *iccmG;
	
	iccmG = &iccmGlob;
	s = Spl(7);						/*	Install our bus error handler	*/
	iccmG->iBusErr = *(long *)8;
	*(long *)8 = ReadError;
	Spl(s);
	
	for(slot=0; slot < ic_SlotMax; slot++)
	{
		iccmG->iCont = nil;
		if (slot == iccmG->iSlot)		/*	if our slot	*/
			continue;
			
		if ((slot == nil) && (iccmArea = (char *)gCom->gMainICCM))				/*	if motherboard and set up	*/
		{
			otherCA = (struct ca_Rec *)MapNuBus(iccmArea);
			if ((ReadLong(&otherCA->Flag) == ~ic_FVal) && (iccmG->iCont == nil))	/*	if valid and no bus error	*/
			{
				iccmG->iCA.Addrs[slot] = iccmArea;
				otherCA->Addrs[iccmG->iSlot] = gCom->gCAP.iPointer;
			}
			continue;
		}
		
		if (TestSlot( slot, &boardID, &lenMem, &iccmArea, CPURsrc, networkRsrc) == false)	/* if can't run A/ROSE	*/
			continue;
		
		if (slot == nil)		/*	motherboard	*/
			iccmEnd = iccmArea + 0x100000;		/*	Check first meg of memory	*/
		else
			iccmEnd = iccmArea + 0x40000;		/*	Check first 128K	*/
		
		while(iccmArea < iccmEnd)
		{
			gcap = (struct gCAP *)MapNuBus( iccmArea );
			if ((ReadLong(&gcap->iMagic) == ~ic_FVal) &&
				(ReadLong(&gcap->iMagic) + ReadLong(&gcap->iPointer) == ReadLong(&gcap->iCksum)) &&
				(iccmG->iCont == nil))
			{
				rawca = (struct ca_Rec *)ReadLong(&gcap->iPointer);
				otherCA = (struct ca_Rec *)MapNuBus( (char *)rawca );
				if ((ReadLong(&otherCA->Flag) == ~ic_FVal) && (iccmG->iCont == nil))		/*	if found and no bus error	*/
					{
						iccmG->iCA.Addrs[slot] = rawca;
						otherCA->Addrs[iccmG->iSlot] = gCom->gCAP.iPointer;
						break;
					}
				}
			if (iccmG->iCont != nil)
				break;
			iccmArea += 2;
		}
	}
	*(long *)8 = iccmG->iBusErr;			/*	Restore bus error handler	*/
}


/********************************************************************************/
/*																				*/
/*		ICCM_Idle	-	Idle chain Scanner routine.								*/
/*																				*/
/*		This routine is used to call the Scanner from the Idle chain.  This		*/
/*		routine is linked into the chain from ICCM_Init routine.				*/
/*																				*/
/*		Entry	none.															*/
/*																				*/
/*		Exit	Scanner called;  next idle task called.							*/
/*																				*/
/********************************************************************************/

void 
ICCM_Idle()
{	
	register struct	iccmGlobals *iccmG;

	iccmG = &iccmGlob;
	iccmG->scanagn = 0xff;
	Scanner();
	iccmG->iIdle();
}


/********************************************************************************/
/*																				*/
/*		ICCM_Init	-	High-level ICCM initialization.							*/
/*																				*/
/*		This routine does the high-level initialization for ICCM.				*/
/*																				*/
/*		Entry	none.															*/
/*																				*/
/*		Exit	Initialization complete.										*/
/*																				*/
/********************************************************************************/

void
ICCM_Init()
{
	message	*msg;
	register struct	iccmGlobals *iccmG;

	iccmG = &iccmGlob;
	
	Reschedule( OS_BLOCK_MODE );		/*	Enter Block mode	*/
	
	iccmG->iSlot = GetCard();					/*	Get card information	*/
	iccmG->iSlot4 = iccmG->iSlot * 4;
	iccmG->iHigh = 0xF0 | iccmG->iSlot;
	
	InitCA(true);							/*	Initialize communications area	*/
	
	ICCReg();							/*	Register ICCM with name manager	*/
	
	FindCards();						/*	Find other cards	*/
	
/*	Set up message for outgoing slot ready.	*/

	while((msg = GetMsg()) == nil)		/*	Get a message	*/
		;
		
	iccmG->iRMP = msg;					/*	Set ready message pointer	*/
	
	msg->mCode = ic_Rdy;				/*	Fill in message fields	*/
	msg->mTo = msg->mFrom;
	
	iccmG->iRdyMsg = msg;				/*	Make message available to scanner	*/

	iccmG->iIdle = gCom->gIdleChain;			/*	Link into idle call chain	*/
	gCom->gIdleChain = ICCM_Idle;

	iccmG->iTick = gCom->gTickChain;			/*	Link into major tick call chain	*/
	gCom->gTickChain = ICCM_Tick;
}


/********************************************************************************/
/*																				*/
/*		ICCM_Tick	-	Tick chain Scanner routine.								*/
/*																				*/
/*		This routine is used to call the Scanner from the Tick chain.  This		*/
/*		routine is linked into the chain from ICCM_Init routine.				*/
/*																				*/
/*		Entry	none.															*/
/*																				*/
/*		Exit	Scanner called;  next tick task called.							*/
/*																				*/
/********************************************************************************/

void
ICCM_Tick()
{	
	register struct	iccmGlobals *iccmG;

	iccmG = &iccmGlob;
	iccmG->scanagn = 0xff;
	Scanner();
	iccmG->iTick();
}


/********************************************************************************/
/*																				*/
/*		InitCA	-	Initialize communications area.								*/
/*																				*/
/*		This routine is called to initialize the communications area.  Once 	*/
/*		this is called, other cards should be able to find us.					*/
/*																				*/
/*		Entry	initial - if true, clear communications area before beginning.	*/
/*																				*/
/*		Exit	Communications area initialized.								*/
/*																				*/
/********************************************************************************/

void
InitCA(short initial)
{
	void	ClrInt();
	char	*c;
	short	i;
	struct IR_REC	*x;
	register struct	iccmGlobals *iccmG;

	iccmG = &iccmGlob;

/*		Initialize communications area	*/

	if (initial)
	{
		c = (char *)&iccmG->iCA;
		for(i=0; i < sizeof(struct ca_Rec); i++)
			c[i] = nil;
	}

	aInitCA();

/*		Set up pointer to comm. area.	*/
		
	iccmG->iCA.IntFlg = 0x80;					/*	Flag we do not support interrupts	*/
		
	if (x = gCom->gIntRecord)					/*	If we support interrupts	*/
	{
		iccmG->iCA.IntPtr = (struct IR_REC *) ToNuBus((char *) x );
		iccmG->iCA.IntFlg = 1;					/*	Flag we do support interrupts	*/
		*(long *)(gCom->gClrVector) = ClrInt;	/*	Set up clear interrupt vector	*/

		x->SIntPtr = (unsigned long)(x->SIntPtr) | (unsigned long)(gCom->gSlotAdd);	/*	Make sure they are	*/
		x->CIntPtr = (unsigned long)(x->CIntPtr) | (unsigned long)(gCom->gSlotAdd);	/*	NuBus addresses	*/
	}

	if (c = (char *) gCom->gMemMap)			/*	If memory map	*/
	{
		iccmG->iCA.MapPtr = (struct AddrMap *) ToNuBus( c );	/*	Save NuBus address	*/
		iccmG->iCA.MapCnt = 1;
		iccmG->iCA.MapChk = 1;				/*	Indicate entry is valid	*/
	}		

	iccmG->iCA.Flag = ~ic_FVal;				/*	Set up flag	*/
	gCom->gLocIMM = &iccmG->iCA;			/*	Local address of ICCM comm. area	*/
	gCom->gCAP.iPointer = (struct ca_Rec *) ToNuBus((char *) &iccmG->iCA);
	gCom->gCAP.iMagic = ic_FVal;
	gCom->gCAP.iCksum =  (unsigned long)(gCom->gCAP.iMagic) + (unsigned long)(gCom->gCAP.iPointer);	/*	Compute Checksum	*/

	iccmG->TimerInc = GetTickPS() * ic_TMOValue;			/*	Set timeout increment value	*/
	iccmG->Timer	= nil;									/*	Initialize timer for next check of timeout of slots 	*/

	iccmG->iCA.gDebugOn = gCom->gDebugOn; 				/*	Copy debug flag			*/
}




/********************************************************************************/
/*																				*/
/*		POQ	-	Process outbound queues.										*/
/*																				*/
/*		This routine scans the outbound queues for each slot and copies the		*/
/*		messages in the queue if the slot is ready to accept them.  If the		*/
/*		slot is not ready, the routine informs the other slot that it has		*/
/*		work to do by setting the IFlag on the other slot.						*/
/*																				*/
/*		Entry	none.															*/
/*																				*/
/*		Exit	Messages pushed to other slots, if possible.					*/
/*																				*/
/********************************************************************************/

void
POQ()
{
	struct ca_Rec	*OtherCA;		/* Storage for communication area */
	short	slot;
	message	*m, *n;
	short	newWaitMask;
	register struct	iccmGlobals *iccmG;

	iccmG = &iccmGlob;
	
	newWaitMask = nil;
 
	for( slot = 0; slot < ic_SlotMax; ++slot)
	{		
		if((m = iccmG->iHeads[slot]) == nil)	/*	If no messages destined for this slot	*/
			continue;


			
		OtherCA = (struct ca_Rec *)MapNuBus((char *)iccmG->iCA.Addrs[slot]);
		if((iccmG->iCA.Addrs[slot] == nil) || (OtherCA->Flag != ~ic_FVal))	/*	If communications area pointer gone	*/
		{			
			while(m = iccmG->iHeads[slot])		/*	While there are messages to return	*/
			{
				iccmG->iHeads[slot] = m->mNext;				
				if(m->mStatus == 0x8000)	/*	If already been returned once	*/
					FreeMsg(m);
				else
				{					
					m->mCode |= 0x8000;
					m->mStatus = 0x8000;
					SwapTID(m);
					Send(m);
					iccmG->iRCount++;
				}
			}
			continue;
		}
			
		if(OtherCA->IFlags[iccmG->iSlot])		/*	If other slot has control of CA area	*/
		{				
			OtherCA->IFlag = 0xff;				/*	Indicate other slot has work to do	*/
			continue;
		}
			
		if(OtherCA->Ptrs[iccmG->iSlot]  == nil)	/*	If a message is not available	*/
		{
			OtherCA->IFlags[iccmG->iSlot] = 0xff;
			OtherCA->IFlag = 0xff;
			IntCard(slot);
			continue;
		}
			
/*	If message available	*/

		m->mDataPtr = ToNuBus(m->mDataPtr);
		n = (message *)MapNuBus((char *)(OtherCA->Ptrs[iccmG->iSlot]));
		n->mId = m->mId;
		n->mCode = m->mCode;
		n->mStatus = m->mStatus;
		n->mPriority = m->mPriority;
		n->mFrom = m->mFrom;
		n->mTo = m->mTo;
		n->mSData[0] = m->mSData[0];
		n->mSData[1] = m->mSData[1];
		n->mSData[2] = m->mSData[2];
		n->mOData[0] = m->mOData[0];
		n->mOData[1] = m->mOData[1];
		n->mOData[2] = m->mOData[2];
		n->mDataSize = m->mDataSize;
		n->mDataPtr = m->mDataPtr;
		
/*		BlockMove((char *)m + 4, MapNuBus((char *)(OtherCA->Ptrs[iccmG->iSlot]) + 4), 	*/
/*				sizeof(message) - 4 );	*/
		OtherCA = (struct ca_Rec *)MapNuBus((char *)iccmG->iCA.Addrs[slot]);
		OtherCA->IFlags[iccmG->iSlot] = 0xff;
		OtherCA->IFlag = 0xff;
		IntCard(slot);						/*	Interrupt the slot if possible	*/
		iccmG->iOCount ++;
			
		if(iccmG->iHeads[slot] = m->mNext)
			newWaitMask |= 1<< slot;		/*	If another message queued up indicate still waiting	*/
		FreeMsg(m);
	}
	iccmG->iWaitMask = newWaitMask;			/* Store new wait mask */
}


/********************************************************************************/
/*																				*/
/*		Q2Slot - queue message to slot.											*/
/*																				*/
/*		This routine is called to queue a message to a remote slot.  If the		*/
/*		slot is able to accept messages immediately, the message is delivered	*/
/*		to the slot.  If not, the message is placed in the outbound queue for	*/
/*		the slot and POQ will send the message later.							*/
/*																				*/
/*		Entry	none.															*/
/*																				*/
/*		Exit	Message sent or queued.											*/
/*																				*/
/********************************************************************************/

void
Q2Slot(message *m)
{
	struct ca_Rec	*OtherCA;		/*	Storage for communication area	*/
	short	slot;
	register struct	iccmGlobals *iccmG;
	void	(*GW) ();
	message	*n;
	
	iccmG = &iccmGlob;
	
	
	slot = (m->mTo >> 24) & 0xff;			/*	Get slot number where message is to be sent	*/
	
	if(slot & 0xf0)		/*	If an inter-machine manager TID, use other slot number	*/
	{		
		slot = ((~slot >> 4) & 0x0f);
		if(slot == iccmG->iSlot)		/*	If for imm gateway task on this card	*/
		{
			/* cbo - Moved test for 0xF0000000 and IMM_NetCopy above test for
					 IMM present. */
			if (((m->mTo & 0x0FFFFFFF) == nil) && (m->mCode == IMM_NetCopy))	/*	If NetCopy request	*/
			{
				m->mStatus = NetCopy( m->mOData[0], (char *) m->mOData[1], m->mOData[2], (char *)m->mDataPtr, m->mDataSize );
				m->mCode |= 1;
				SwapTID( m );
				Send( m );
				return;
			}
			
			if(GW = GetgCommon()->gIMMHandler)		/*	If gateway task present	*/
			{
				GW(m);
				return;
			}
			
			if (m->mStatus == 0x8000)		/*	If previously undeliverable	*/
			{	
				FreeMsg( m );
				return;
			}
			
			m->mCode |= 0x8000;				/*	Return to sender as undeliverable	*/
			m->mStatus = 0x8000;
			SwapTID(m);
			Send(m);
			iccmG->iRCount++;
			return;
		}
	}
	
	if(iccmG->iWaitMask & (1 << slot))		/*	If message already queued to slot, queue this message as well	*/
	{
		m->mNext = 0;
		iccmG->iTails[slot]->mNext = m;
		iccmG->iTails[slot] = m;
		iccmG->iQCount++;
		return;
	}
	
	OtherCA = (struct ca_Rec *)MapNuBus((char *)iccmG->iCA.Addrs[slot]);
	if((iccmG->iCA.Addrs[slot]) && (OtherCA->Flag == ~ic_FVal))		/*	If slot is active 	*/
	{		
		if(OtherCA->IFlags[iccmG->iSlot])		/*	If other slot has control of CA area	*/
		{				
			OtherCA->IFlag = 0xff;		/*	Indicate other slot has work to do	*/
			m->mNext = 0;				/*	Link message into wait queue	*/
			iccmG->iHeads[slot] = m;
			iccmG->iTails[slot] = m;
			iccmG->iWaitMask |= 1 << slot;
			iccmG->iQCount++;
			return;
		}
		
		if(OtherCA->Ptrs[iccmG->iSlot])		/*	If a message is available	*/
		{
			m->mDataPtr = ToNuBus(m->mDataPtr);

			n = (message *)MapNuBus((char *)(OtherCA->Ptrs[iccmG->iSlot]));			
			n->mId = m->mId;
			n->mCode = m->mCode;
			n->mStatus = m->mStatus;
			n->mPriority = m->mPriority;
			n->mFrom = m->mFrom;
			n->mTo = m->mTo;
			n->mSData[0] = m->mSData[0];
			n->mSData[1] = m->mSData[1];
			n->mSData[2] = m->mSData[2];
			n->mOData[0] = m->mOData[0];
			n->mOData[1] = m->mOData[1];
			n->mOData[2] = m->mOData[2];
			n->mDataSize = m->mDataSize;
			n->mDataPtr = m->mDataPtr;
/*			BlockMove((char *) m + 4, MapNuBus((char *)(OtherCA->Ptrs[iccmG->iSlot]) + 4), 	*/
/*					sizeof(message) - 4);	*/
			FreeMsg(m);
			
			OtherCA = (struct ca_Rec *)MapNuBus((char *)iccmG->iCA.Addrs[slot]);
			OtherCA->IFlags[iccmG->iSlot] = 0xff;
			OtherCA->IFlag = 0xff;
			IntCard(slot);				/*	Interrupt the slot if possible	*/
			iccmG->iOCount ++;
			return;
		}

		/* Message is not available */
		
		OtherCA->IFlags[iccmG->iSlot] = 0xff;
		OtherCA->IFlag = 0xff;
		IntCard(slot);
		
		m->mNext = 0;					/*	Link message into wait queue	*/
		iccmG->iHeads[slot] = m;
		iccmG->iTails[slot] = m;
		iccmG->iWaitMask |= 1 << slot;
		iccmG->iQCount++;
		return;
	}
	
/*	Message cannot be sent to slot - return as undeliverable	*/
			
	if(m->mStatus == 0x8000)		/*	If already been returned once	*/
	{		
		FreeMsg(m);
		return;
	}
		
	m->mCode |= 0x8000;		/*	Return message as undeliverable	*/
	m->mStatus = 0x8000;
	SwapTID(m);
	Send(m);
	iccmG->iRCount++;
}




/********************************************************************************/
/*																				*/
/*		RCI - return card information.											*/
/*																				*/
/*		This routine returns information about the slots on this machine.  If	*/
/*		card information is requested (ICC_GETCARDS), the TID of the name		*/
/*		manager is returned for each slot.  If IMM information is requested		*/
/*		(ICC_GETIMMS), the IMM manager TID for each slot is returned.			*/
/*																				*/
/*		Entry	msg 	-	pointer to request message.							*/
/*																				*/
/*		Exit	Reply buffer filled in and message returned.					*/
/*																				*/
/********************************************************************************/

void
RCI(message *msg)

{
	tid_type	temptid;
	short		slot;
	
	struct ca_Rec	*otherCA;		/* Storage for communication area */
	struct	ra_GetCards *icptr;
	struct	ra_GetIMMs *imptr;
	register struct	iccmGlobals *iccmG;

	iccmG = &iccmGlob;
	
	switch (msg->mCode)
	{
		case ICC_GETCARDS:
		
			if((sizeof(struct ra_GetCards) >= msg->mDataSize) && (msg->mDataPtr != nil))
			{				
				for(slot = ic_SlotMax-1; slot >= 0; slot--)
				{
					icptr = (struct	ra_GetCards *)MapNuBus(msg->mDataPtr);
					icptr->tid[slot] = -1;				/* Preset slot not active */
					
					if(iccmG->iCA.Addrs[slot])		/** If slot is active	*/
					{
						otherCA = (struct ca_Rec *)MapNuBus((char *)iccmG->iCA.Addrs[slot]);
						if(otherCA->Flag == ~ic_FVal)	/*	If other communications area is still set up correctly	*/
						{
							temptid = otherCA->NameMgr;
							icptr = (struct	ra_GetCards *)MapNuBus(msg->mDataPtr);
							icptr->tid[slot] = temptid;	/* Return Name Manager TID or zero	*/
						}
						else
							iccmG->iCA.Addrs[slot] = 0;  /*	If other communications area is not up, indicate inactive	*/
					}
				}
				icptr = (struct	ra_GetCards *)MapNuBus(msg->mDataPtr);
				icptr->tid[iccmG->iSlot] = gCom->gNameTask;	/* Return my Name Manager TID	*/
			}
			else
				msg->mStatus = ic_BadReplyArea;		/*	If return area is not large enough for reply	*/
			break;
			
		case ICC_GETIMMS:
		
			if((sizeof(struct ra_GetIMMs) >= msg->mDataSize) && (msg->mDataPtr != 0))
			{				
				for(slot = ic_SlotMax-1; slot >= 0; slot--)
				{
					imptr = (struct	ra_GetIMMs *)MapNuBus(msg->mDataPtr);
					imptr->tid[slot] = -1;		/*	Preset slot not active	*/
					
					if(iccmG->iCA.Addrs[slot])		/*	If slot is active	*/
					{
						otherCA = (struct ca_Rec *)MapNuBus((char *)iccmG->iCA.Addrs[slot]);
						if(otherCA->Flag == ~ic_FVal)		/*	If other communications area is still set up correctly	*/
						{
							temptid = otherCA->NameMgr;
							imptr = (struct	ra_GetIMMs *)MapNuBus(msg->mDataPtr);
							imptr->tid[slot] = temptid;		/*	Return IMM TID or zero	*/
						}
						else
							iccmG->iCA.Addrs[slot] = 0;		/*	If other communications area is not up, indicate inactive	*/
					}
				}
				imptr = (struct	ra_GetIMMs *)MapNuBus(msg->mDataPtr);
				imptr->tid[iccmG->iSlot] = gCom->gIMMTask;	/* Return my IMM TID	*/
			}
			else
				msg->mStatus = ic_BadReplyArea;		/*	If return area is not large enough for reply	*/
			break;
			
		default:
			msg->mStatus = OS_UNKNOWN_MESSAGE;		/*	Request unrecognized	*/
			break;
	}
	SwapTID( msg );
	msg->mCode |= 1;
	Send(msg);
}


/*******************************************************************************/
/*																				*/
/*		Scan	-	Scan slot for work to do.									*/
/*																				*/
/*		This routine is used to check for incoming messages from a given slot.	*/
/*		If there is a message:													*/
/*			A new message buffer is allocated.									*/
/*			The message received is sent to the destination task.				*/
/*																				*/
/*		Entry	slot	-	slot number.										*/
/*																				*/
/*		Exit	Communications area initialized.								*/
/*																				*/
/********************************************************************************/

void 
Scan(slot)
short	slot;
{
	message	*m;
	register struct	iccmGlobals *iccmG;

	iccmG = &iccmGlob;
	
	m = iccmG->iCA.Ptrs[slot];
	iccmG->iCA.Ptrs[slot] = (struct mMessage *)ToNuBus((char *)GetMsg());
	
	if(iccmG->iCA.Ptrs[slot])		/*	If I got a message buffer for this slot	*/
		iccmG->iCA.IFlags[slot] = 0;
	else		
		iccmG->iCA.IFlag = 0xff;	/*	Indicate a slot still needs a buffer	*/
	
	if(m)	/*	If we got an incoming message	*/
	{		
		iccmG->iICount++;
		Send(m);			/*	Send the message	*/
	}
}

