/*================================================================================
	reportError.c
	version 1.1
	
	Greg Anderson
	28 June 1991
	greggor@apple.com
	
	These routines are very convenient for debugging; they will convert
	OSErrors from short negative numbers to human-readable strings in
	Pascal format
	
	This code depends on:
	
		stringUtilities.c
		dialogUtilities.c
		
		'DLOG' resource ID 30303	A simple dialog box for displaying errors;
		'DITL' resource ID 30303	Has one 'Okay' button and one static text
									item that reads:
									
									^0 ^1
									
									^2 because ^3
									
		'TEXT' resource ID 30303	All error messages, each line cr-terminated
		
================================================================================*/
#include "reportError.h"

/*
// Header files for utility routines:
*/
#ifndef __DIALOGUTILITIES__
	#include "dialogUtilities.h"
#endif
#ifndef __STRINGUTILITIES__
	#include "stringUtilities.h"
#endif

/*
// Macintosh includes
*/
#ifndef __TYPES__
	#include <Types.h>
#endif
#ifndef __ERRORS__
	#include <Errors.h>
#endif
#ifndef __MEMORY__
	#include <Memory.h>
#endif
#ifndef __RESOURCES__
	#include <Resources.h>
#endif
#ifndef __PACKAGES__
	#include <Packages.h>
#endif


Handle		resultCodeText = nil;

/*--------------------------------------------------------------------------------
	Search for an error result code in the result code text
	
	Return 'false' if the name and description strings cannot be filled in
--------------------------------------------------------------------------------*/
pascal Boolean FindResultCodeDescription( OSErr theErr, Str255 errorNameStr, Str255 errorDescStr )
{
	Size		textSize;
	char*		p;
	Boolean		foundEntry;
	OSErr		testError;
	
	/*
	// Note that if the error is a 'memFullErr', it isn't going to do
	// much good to try to load in the result code test.  Handle this
	// case separately.
	//
	// While there are other situations that could also cause this
	// routine to fail, none are as common as memFullErr.
	*/
	if( theErr == memFullErr )
	{
		CtoPcpy( errorNameStr, "memFullErr" );
		CtoPcpy( errorDescStr, "there is insufficient memory left in the heap zone" );
		
		return true;
	}
	/*
	// If we have never loaded the result code text, do so now
	*/
	if( resultCodeText == nil )
		resultCodeText = GetResource( 'TEXT', kReportErrorID );
	if( resultCodeText == nil )
		return false;
	/*
	// If the result code text has been purged, try to re-load it.
	*/
	if( *resultCodeText == nil )
		LoadResource( resultCodeText );
	if( *resultCodeText == nil )
		return false;
	/*
	// Lock the result text & dereference it
	*/
	MoveHHi( resultCodeText );
	HLock( resultCodeText );
	p = *resultCodeText;
	textSize = GetHandleSize( resultCodeText );
	foundEntry = false;
	do
	{
		SkipToSpec( p, '-' );
		if( !(*p) )
			break;
		++p;
		
		testError = -ScanNumberInString( &p );
		if( testError == theErr )
		{
			SkipPastWhitespace( p );
			ScanWordInString( &p, errorNameStr );
			SkipPastWhitespace( p );
			ScanLineInString( &p, errorDescStr );
			foundEntry = true;
			break;
		}
	}
	while( testError != -32767 );
	/*
	// Unlock the result text and return
	*/
	HUnlock( resultCodeText );
	return foundEntry;
}

/*--------------------------------------------------------------------------------
	Put up a dialog box reporting an error to the user
--------------------------------------------------------------------------------*/
pascal void ReportError( Str255 errorMsg, OSErr theErr )
{
	DialogPtr		dlog;
	short			itemHit;
	Str255			errorNumberStr;
	Str255			errorNameStr;
	Str255			errorDescStr;
	long			longErr = theErr;
	
	/*
	// Do nothing if theErr == noErr
	*/
	if( theErr != noErr )
	{
		/*
		// Clear out the pascal strings, just to be safe
		*/
		errorNameStr[0] = 0;
		errorDescStr[0] = 0;
		errorNumberStr[0] = 0;
		/*
		// Set up the error number string
		*/
		NumToString( longErr, errorNumberStr );
		FindResultCodeDescription( theErr, errorNameStr, errorDescStr );
		if( errorDescStr[0] == 0 )
			CtoPcpy( errorDescStr, "an error ocurred" );
		/*
		// Set up Dialog box
		*/
		dlog = GetNewDialog(kReportErrorID, (Ptr)0L, (WindowPtr)-1L);
		if( dlog != nil )
		{
			SetPort( dlog );
			ParamText( errorMsg, errorNumberStr, errorNameStr, errorDescStr );
			SelectWindow(dlog);
			InstallDefaultOutline( dlog, 1 );
			CenterAndShowDialog(dlog);
						
			/*
			// Wait for user to click "Okay"
			*/
			do
			{
				ModalDialog((ModalFilterProcPtr) CutPasteFilter, &itemHit);
			} while( itemHit > 2 );
			DisposDialog(dlog);
		}
		else
		{
			DebugStr( (unsigned char*)"\pReportError could not bring up dialog box" );
		}
	}
}
